#include <vector>
#include <algorithm>
#include <map>

enum {
  UNASSIGNED=-1,
  ELIMINATED=-2
};

void increaseElement(std::vector<unsigned int>& vector,unsigned int element)
{
  if (element<vector.size())
    ++vector[element];
  else {
    vector.resize(element+1);
    vector[element]=1;
  }
}

class Schedule;

class Fixed {
  unsigned int numPlayers,remainingPlayers;
  int lives;
  std::vector<unsigned int> plays;
  std::vector<int> losses,playerToRank;
  unsigned int maxEqualLossScore;
  std::vector<std::vector<unsigned int> > pairings;

  friend class Schedule;

  Fixed(const std::vector<std::vector<unsigned int> >& defeats,const int lives_)
  : numPlayers(defeats.size())
  , lives(lives_)
  , plays(std::vector<unsigned int>(numPlayers))
  , losses(std::vector<int>(numPlayers))
  , playerToRank(std::vector<int>(numPlayers))
  , pairings(defeats)
  {
    for (unsigned int winner=0;winner<numPlayers;++winner)
      for (unsigned int loser=0;loser<numPlayers;++loser) {
        plays[winner]+=defeats[winner][loser];
        plays[loser ]+=defeats[winner][loser];
        losses[loser]+=defeats[winner][loser];
        pairings[loser][winner]+=defeats[winner][loser];
      }
    std::vector<std::pair<unsigned int,unsigned int> > lossPlayerPairs;
    lossPlayerPairs.reserve(numPlayers);
    for (unsigned int player1=0;player1<numPlayers;++player1)
      if (losses[player1]<lives)
        lossPlayerPairs.push_back(std::make_pair(losses[player1],player1));
    remainingPlayers=lossPlayerPairs.size();
    /*stable_*/sort(lossPlayerPairs.begin(),lossPlayerPairs.end());
    unsigned int maxEqualLossDiff=0;
    unsigned int equalLossDiff=0;
    for (unsigned int rank=0;rank<remainingPlayers;++rank) {
      const std::pair<unsigned int,unsigned int>& lossPlayerPair=lossPlayerPairs[rank];
      playerToRank[lossPlayerPair.second]=rank;
      if (rank==remainingPlayers-1 || lossPlayerPair.first!=lossPlayerPairs[rank+1].first) {
        if (equalLossDiff>maxEqualLossDiff)
          maxEqualLossDiff=equalLossDiff;
        equalLossDiff=0;
      }
      else
        ++equalLossDiff;
    }
    maxEqualLossScore=maxEqualLossDiff*maxEqualLossDiff;
  }
};

class Schedule {
  const Fixed* f;
  int bye;
  std::vector<unsigned int> nthTimePairings,nLossDiffPairings;
  unsigned int score;
  std::vector<int> pairedWith;

  Schedule() : f(0) {}
  Schedule(const Schedule& original,unsigned int exemptedPlayer)
  : f(original.f)
  , bye(exemptedPlayer)
  , nthTimePairings(original.nthTimePairings)
  , nLossDiffPairings(original.nLossDiffPairings)
  , score(original.score)
  , pairedWith(original.pairedWith)
  {
    pairedWith[exemptedPlayer]=exemptedPlayer;
  }
  Schedule(const Schedule& original,unsigned int player1,unsigned int player2)
  : f(original.f)
  , bye(original.bye)
  , nthTimePairings(original.nthTimePairings)
  , nLossDiffPairings(original.nLossDiffPairings)
  , score(original.score)
  , pairedWith(original.pairedWith)
  {
    pair(player1,player2);
  }
  void pair(unsigned int player1,unsigned int player2)
  {
    pairedWith[player1]=player2;
    pairedWith[player2]=player1;
    const unsigned int occurrences=f->pairings[player1][player2];
    --nthTimePairings[occurrences];
    increaseElement(nthTimePairings,occurrences+1);
    const unsigned int lossDifference=abs(f->losses[player1]-f->losses[player2]);
    increaseElement(nLossDiffPairings,lossDifference);
    const int rankDifference=f->playerToRank[player1]-f->playerToRank[player2];
    if (lossDifference==0)
      score+=f->maxEqualLossScore-rankDifference*rankDifference;
    else
      score+=rankDifference*rankDifference;
  }
  Schedule(const Fixed& f_)
  : f(&f_)
  , bye(UNASSIGNED)
  , score(0)
  , pairedWith(std::vector<int>(f->numPlayers,UNASSIGNED))
  {
    for (unsigned int player1=0;player1<f->numPlayers;++player1) {
      for (unsigned int player2=player1+1;player2<f->numPlayers;++player2)
        increaseElement(nthTimePairings,f->pairings[player1][player2]);
      if (f->losses[player1]>=f->lives)
        pairedWith[player1]=ELIMINATED;
    }
  }

  static void best(Schedule& bestSchedule,const Schedule& currentSchedule)
  {
    unsigned int player1=0;
    while (true) {
      if (player1<currentSchedule.f->numPlayers) {
        if (currentSchedule.pairedWith[player1]==UNASSIGNED)
          break;
        ++player1;
      }
      else {
        if (currentSchedule<bestSchedule)
          bestSchedule=currentSchedule;
        return;
      }
    }
    std::vector<Schedule> choices;
    choices.reserve(currentSchedule.f->remainingPlayers-1);
    for (unsigned int player2=player1+1;player2<currentSchedule.f->numPlayers;++player2)
      if (currentSchedule.pairedWith[player2]==UNASSIGNED) {
        const Schedule candidate(Schedule(currentSchedule,player1,player2));
        if (bestSchedule<candidate)
          continue;
        else
          choices.push_back(candidate);
      }
    sort(choices.begin(),choices.end());
    for (std::vector<Schedule>::const_iterator choice=choices.begin();choice!=choices.end();++choice)
      if (bestSchedule<*choice)
        return;
      else
        best(bestSchedule,*choice);
  }

  static bool ideal(Schedule& bestSchedule,const Schedule& currentSchedule)
  {
    typedef std::map<unsigned int,std::vector<unsigned int> > Map;
    Map lossToPlayers;
    for (unsigned int player=0;player<currentSchedule.f->numPlayers;++player)
      if (currentSchedule.pairedWith[player]==UNASSIGNED)
        lossToPlayers[currentSchedule.f->losses[player]].push_back(player);
    Schedule candidate(currentSchedule);
    for (Map::const_iterator lossBracket=lossToPlayers.begin();lossBracket!=lossToPlayers.end();++lossBracket) {
      if (lossBracket->second.size()%2==1)
        return false;
      else {
        unsigned int topIndex=0;
        unsigned int bottomIndex=lossBracket->second.size()-1;
        for (;topIndex<bottomIndex;++topIndex,--bottomIndex) {
          const unsigned int top   =lossBracket->second[   topIndex];
          const unsigned int bottom=lossBracket->second[bottomIndex];
          const unsigned int occurrences=currentSchedule.f->pairings[top][bottom];
          for (unsigned int player=0;player<currentSchedule.f->numPlayers;++player) {
            if (currentSchedule.pairedWith[player]==UNASSIGNED &&
                player!=top &&
                player!=bottom &&
                (currentSchedule.f->pairings[player][top   ]<occurrences ||
                 currentSchedule.f->pairings[player][bottom]<occurrences))
              return false;
          }
          candidate.pair(top,bottom);
        }
      }
    }

    if (candidate<bestSchedule)
      bestSchedule=candidate;
    return true;
  }

public:
  static std::vector<int> best(const std::vector<std::vector<unsigned int> >& defeats,const int lives)
  {
    Schedule bestSchedule;
    Fixed fixed(defeats,lives);
    Schedule currentSchedule(fixed);
    if (fixed.remainingPlayers%2==1) {
      // odd number of players left
      std::vector<Schedule> choices;
      choices.reserve(fixed.remainingPlayers);
      for (unsigned int player=0;player<fixed.numPlayers;++player)
        if (currentSchedule.pairedWith[player]==UNASSIGNED)
          choices.push_back(Schedule(currentSchedule,player));
      sort(choices.begin(),choices.end());
      for (std::vector<Schedule>::const_iterator choice=choices.begin();choice!=choices.end();++choice) {
        if (bestSchedule<*choice)
          break;
        else if (!ideal(bestSchedule,*choice))
          best(bestSchedule,*choice);
      }
    }
    else {
      // even number of players left
      if (!ideal(bestSchedule,currentSchedule))
        best(bestSchedule,currentSchedule);
    }

    return bestSchedule.pairedWith;
  }

  bool operator<(const Schedule& rhs) const
  {
    if (f==0 || rhs.f==0)
      return f!=0;
    if (bye!=rhs.bye) {
      /*if (bye==UNASSIGNED || rhs.bye==UNASSIGNED)
        abort();*/
      if (f->plays[bye]!=rhs.f->plays[rhs.bye])
        return f->plays[bye]>rhs.f->plays[rhs.bye];
    }
    for (int n=std::max(nthTimePairings.size(),rhs.nthTimePairings.size())-1;n>=2;--n) {
      const unsigned int left =(static_cast<unsigned int>(n)<    nthTimePairings.size() ?     nthTimePairings[n] : 0);
      const unsigned int right=(static_cast<unsigned int>(n)<rhs.nthTimePairings.size() ? rhs.nthTimePairings[n] : 0);
      if (left!=right)
        return left<right;
    }
    if (bye!=UNASSIGNED && bye!=rhs.bye) {
      if (f->losses[bye]!=rhs.f->losses[rhs.bye])
        return f->losses[bye]<rhs.f->losses[rhs.bye];
      else
        return bye<rhs.bye;
    }
    for (int n=std::max(nLossDiffPairings.size(),rhs.nLossDiffPairings.size())-1;n>=1;--n) {
      const unsigned int left =(static_cast<unsigned int>(n)<    nLossDiffPairings.size() ?     nLossDiffPairings[n] : 0);
      const unsigned int right=(static_cast<unsigned int>(n)<rhs.nLossDiffPairings.size() ? rhs.nLossDiffPairings[n] : 0);
      if (left!=right)
        return left<right;
    }
    return score<rhs.score;
  }
};

#include <cstdlib>
#include <fstream>
#include <sstream>
#include <iostream>
#include "colors.hpp"

int main(int argc, char* argv[])
{
  if (argc<4)
    return EXIT_FAILURE;
  std::string line,goldPlayer,silverPlayer;
  ColorHistory colorHistory;
  for (int arg=1;arg<argc-3;++arg) {
    std::ifstream past(argv[arg]);
    while (std::getline(past,line)) {
      if (line.empty())
        continue;
      std::stringstream ss;
      ss<<line;
      if (!(ss>>goldPlayer))
        continue;
      if (!(ss>>silverPlayer))
        abort();
      colorHistory.addGame(goldPlayer,silverPlayer);
    }
  }
  std::ifstream seed(argv[argc-2]);
  std::vector<std::string> players;
  std::map<std::string,unsigned int> playerToSeed;
  std::string player;
  while (seed>>player) {
    playerToSeed.insert(make_pair(player,players.size()));
    players.push_back(player);
  }
  const int numPlayers=players.size();
  std::vector<std::vector<unsigned int> > defeats(numPlayers,std::vector<unsigned int>(numPlayers));
  std::ifstream tournament(argv[argc-1]);
  while (std::getline(tournament,line)) {
    if (line.empty())
      continue;
    std::stringstream ss;
    ss<<line;
    if (!(ss>>goldPlayer))
      continue;
    if (playerToSeed.find(goldPlayer)==playerToSeed.end())
      abort();
    if (!(ss>>silverPlayer) || playerToSeed.find(silverPlayer)==playerToSeed.end())
      abort();
    std::string winner;
    if (!(ss>>winner) || (winner!=goldPlayer && winner!=silverPlayer))
      abort();
    ++defeats[playerToSeed[winner]][playerToSeed[winner==goldPlayer ? silverPlayer : goldPlayer]];
    colorHistory.addGame(goldPlayer,silverPlayer);
  }
  srand(time(0));
  const std::vector<int> pairings=Schedule::best(defeats,atoi(argv[argc-3]));
  for (int n=0;n<numPlayers;++n)
    if (pairings[n]==n)
      std::cout<<players[n]<<'\n';
    else if (pairings[n]>n /*&& pairings[n]!=ELIMINATED*/) {
      const std::string player1=players[n],player2=players[pairings[n]];
      const int color=colorHistory.colorFirstPlayer(player1,player2);
      if (color>0 || (color==0 && (rand()%2)==0)) {
          goldPlayer=player1;
        silverPlayer=player2;
      }
      else {
        silverPlayer=player1;
          goldPlayer=player2;
      }
      std::cout<<goldPlayer<<'\t'<<silverPlayer<<'\n';
      //colorHistory.addGame(goldPlayer,silverPlayer);
    }
}

